#-*- coding:utf-8 -*-

import json
from django.contrib.auth.models import User
from django.core.cache import cache
from django.db.models import Max
from dictionary.models import Lexeme, LexemeInflectionPattern, PartOfSpeech, \
  Pattern, InflectionCharacteristic, Vocabulary, LexemeAssociation, Qualifier, \
  prepare_table, visible_vocabularies, ClassificationValue, CrossReference, \
  TableTemplate, get_root, InputLexeme, set_vocabulary
from dictionary.forms import LexemeEditForm, LIPEditForm, ClassificationForm, \
  CrossReferenceForm
from common.decorators import render, ajax, AjaxError
from common.util import error_messages

from dictionary.ajax_history import *
from dictionary.ajax_lexeme_jqgrid import *
from dictionary.ajax_prompter import *
from dictionary.ajax_filters import *


@render('inflection_tables.html')
@ajax(method='get', encode_result=False)
def get_inflection_tables(request, variant, id):
  lexeme = Lexeme.objects.get(pk=id)
  if lexeme.owner_vocabulary not in visible_vocabularies(request.user):
    raise AjaxError('access denied')
  qualifiers = Qualifier.objects.filter(
    vocabulary__in=visible_vocabularies(request.user))
  tables = lexeme.inflection_tables(variant, qualifiers=qualifiers)
  return {'tables': tables}

@render('inflection_table.html')
@ajax(method='get', encode_result=False)
def table_preview(request, id, pattern, inflection_characteristic, lip_id,
                  entry=None, pos=None):
  lexeme = Lexeme.objects.get(pk=id)
  if lexeme.owner_vocabulary not in visible_vocabularies(request.user):
    raise AjaxError('access denied')
  if entry is None:
    entry = lexeme.entry
  if pos is None:
    pos = lexeme.part_of_speech.symbol
  try:
    pattern = Pattern.objects.get(name=pattern)
    inflection_characteristic = InflectionCharacteristic.objects.get(
      pk=inflection_characteristic)
    if lip_id.startswith('lip_add'):
      lip = LexemeInflectionPattern(lexeme=lexeme, index=0)
    else:
      lip = LexemeInflectionPattern.objects.get(pk=int(lip_id[3:]))
    lip.pattern = pattern
    lip.inflection_characteristic = inflection_characteristic
    lip.root = get_root(entry, pos, pattern, inflection_characteristic)
    qualifiers = Qualifier.objects.filter(
      vocabulary__in=visible_vocabularies(request.user))
    table = lip.inflection_table('0', separated=True, qualifiers=qualifiers,
                                 edit_view=True)
    prepare_table(table)
  except Pattern.DoesNotExist:
    table = None
  except TableTemplate.DoesNotExist:
    table = None
  return {'table': table}

@render()
@ajax(method='get', encode_result=False)
def odm_forms(request, id):
  to_return = {}
  l = Lexeme.objects.get(pk=id)
  if l.owner_vocabulary not in visible_vocabularies(request.user):
    raise AjaxError('access denied')
  odm_lexemes = []
  for l in InputLexeme.objects.filter(entry=l.entry):
    odm_lexemes.append(list(l.inputform_set.values_list('form', flat=True)))
  to_return['odm_lexemes'] = odm_lexemes
  return to_return

@render()
@ajax(method='get', encode_result=False)
def lexeme_edit_form(request, id):
  to_return = {}
  l = Lexeme.objects.get(pk=id)
  if l.owner_vocabulary not in visible_vocabularies(request.user):
    raise AjaxError('access denied')
  editable = l.perm(request.user, 'change')
  to_return['editable'] = editable
  owner = l.owner_vocabulary
  visible_vocabs = visible_vocabularies(request.user)
  r_vocabularies = visible_vocabs.exclude(pk=owner.pk)
  containing = r_vocabularies.filter(lexemeassociation__lexeme=l)
  ro_vocabularies = containing.exclude(editors=request.user)
  rw_vocabularies = r_vocabularies.filter(editors=request.user)
  all_rw_vocabularies = l.rw_vocabularies(request.user)
  #all_ro_vocabularies = visible_vocabs.exclude(editors=request.user)
  ro_owner = owner not in all_rw_vocabularies
  to_return['multiple_editable'] = editable and bool(rw_vocabularies)
  if ro_owner:
    ro_qualifiers = l.qualifiers.filter(vocabulary=owner) #__in=all_ro_vocabularies)
  else:
    ro_qualifiers = []
  to_return['owner'] = owner
  to_return['ro_vocabularies'] = ro_vocabularies
  to_return['ro_qualifiers'] = ro_qualifiers

  to_return['form'] = LexemeEditForm(
    instance=l, editable=editable, vocabularies=rw_vocabularies,
    containing=rw_vocabularies.filter(
      pk__in=containing.values_list('pk', flat=True)),
    rw_vocabularies=all_rw_vocabularies,
    all_rw_vocabularies=request.user.editable_vocabularies.all())
  to_return['id'] = l.pk
  to_return['part_of_speech'] = l.part_of_speech.symbol
  to_return['classification_forms'] = make_classification_forms(
    l, owner, editable);

  lips = l.lexemeinflectionpattern_set.all()
  to_return['lip_forms'] = [
    (LIPEditForm(
      part_of_speech=l.part_of_speech, instance=lip, prefix='lip' + str(lip.pk),
      rw_vocabularies=all_rw_vocabularies, editable=editable),
     lip.qualifiers.filter(vocabulary=owner) if ro_owner else [])
    for lip in lips]
  crs = l.refs_to.filter(to_lexeme__deleted=False).order_by('type__index')
  to_return['cross_references'] = crs
  return to_return

def make_classification_forms(lexeme, vocabulary=None, editable=True):
  if not vocabulary:
    vocabulary = lexeme.owner_vocabulary
  classifications = vocabulary.classifications.all()
  classification_forms = []
  for c in classifications:
    values = lexeme.classification_values(c)
    classification_forms.append(
      ClassificationForm(
        classification=c, values=values, prefix='cl' + str(c.pk),
        editable=editable))
  return classification_forms

@render()
@ajax(method='get', encode_result=False)
def classification_forms(request, lexeme_id, vocab_id):
  l = Lexeme.objects.get(pk=lexeme_id)
  if vocab_id:
    vocab = Vocabulary.objects.get(pk=vocab_id)
  else:
    vocab = None
  forms = make_classification_forms(l, vocab)
  return {'forms': forms}

@render('lexeme_edit_form_row.html')
@ajax(method='get', encode_result=False)
def new_lip_edit_row(request, id, pos_id):
  l = Lexeme.objects.get(pk=id)
  if l.owner_vocabulary not in request.user.editable_vocabularies.all():
    raise AjaxError('access denied')
  if not pos_id:
    raise AjaxError(u'Nieokreślona część mowy.')
  pos = PartOfSpeech.objects.get(pk=pos_id)
  rw_vocabularies = l.rw_vocabularies(request.user)
  lip_form = LIPEditForm(part_of_speech=pos, prefix='lip_add_NUM',
                         rw_vocabularies=rw_vocabularies)
  return {'lip_form': lip_form, 'editable': True}

@render('cross_reference_row.html')
@ajax(method='get', encode_result=False)
def new_cross_reference_row(request, id, pos_id):
  l = Lexeme.objects.get(pk=id)
  if l.owner_vocabulary not in request.user.editable_vocabularies.all():
    raise AjaxError('access denied')
  pos = PartOfSpeech.objects.get(pk=pos_id)
  cr_form = CrossReferenceForm(lexeme=l, pos=pos, prefix='cr_add_NUM')
  return {'cr_form': cr_form, 'editable': True}

@ajax(method='post')
def update_lexeme(request, form_data, mask=''):
  form_dict = dict((x['name'], x['value']) for x in form_data)
  l = Lexeme.objects.get(pk=form_dict['id'])
  if not l.perm(request.user, 'view'):
    raise AjaxError('access denied')
  owner = l.owner_vocabulary
  created = l.entry == ''
  visible_vocabs = visible_vocabularies(request.user)
  r_vocabularies = visible_vocabs.exclude(pk=owner.pk)
  rw_vocabularies = r_vocabularies.filter(editors=request.user)
  all_rw_vocabularies = l.rw_vocabularies(request.user)
  if not l.perm(request.user, 'change'):
    return update_lexeme_qualifiers(
      l, all_rw_vocabularies, rw_vocabularies, form_dict, form_data)
  form = LexemeEditForm(
    data=form_dict, instance=l, vocabularies=rw_vocabularies,
    rw_vocabularies=all_rw_vocabularies,
    all_rw_vocabularies=request.user.editable_vocabularies.all())
  if form.is_valid():
    l = form.save()
    l.responsible = request.user
    l.deleted = False
    l.save()
  else:
    raise AjaxError(error_messages(form))

  for vocab in rw_vocabularies:
    set_vocabulary(l, vocab, vocab in form.cleaned_data['vocabularies'])
  new_owner = form.cleaned_data.get('new_owner')
  if new_owner and new_owner != owner:
    set_vocabulary(l, new_owner, True)
    l.owner_vocabulary = new_owner
    set_vocabulary(l, owner, False)
    l.save()
  for qualifier in form.fields['qualifiers'].queryset:
    qualifier.set_for(l, qualifier in form.cleaned_data['qualifiers'])

  classifications = l.owner_vocabulary.classifications.all()
  for c in classifications:
    classification_form = ClassificationForm(
      data=form_dict, classification=c, prefix='cl' + str(c.pk))
    if classification_form.is_valid():
      cvs = ClassificationValue.objects.filter(
        pk__in=classification_form.cleaned_data['values'])
      l_cvs = l.classification_values(c)
      for cv in l_cvs:
        if cv.pk not in classification_form.cleaned_data['values']:
          cv.lexemes.remove(l) #add
      for cv in cvs:
        if cv not in l_cvs:
          cv.lexemes.add(l) #add
    else:
      raise AjaxError(error_messages(classification_form))
  for cv in l.classificationvalue_set.all():
    if cv.classification not in classifications:
      cv.lexemes.remove(l) #add

  submitted_lips = []
  submitted_crs = []
  for prefix in form_dict['deleted']:
    pk = int(prefix[3:])
    LexemeInflectionPattern.objects.get(pk=pk).delete()
  for cr_pk in form_dict['deleted_cr']:
    CrossReference.objects.get(pk=cr_pk).delete()
  for pair in form_data:
    # może (?) się psuć, jeśli jest zły charfl
    name = pair['name']
    prefix = name.split('-')[0]
    if name.startswith('lip') and prefix not in submitted_lips:
      submitted_lips.append(prefix)
      if prefix.startswith('lip_add'):
        lip = LexemeInflectionPattern()
        lip.lexeme = l
      else:
        pk = int(prefix[3:])
        lip = LexemeInflectionPattern.objects.get(pk=pk)
      form_dict[prefix + '-qualifiers'] = get_list(
        form_data, prefix + '-qualifiers')
      lip_form = LIPEditForm(
        part_of_speech=l.part_of_speech, data=form_dict, prefix=prefix,
        instance=lip, rw_vocabularies=all_rw_vocabularies,
        index=len(submitted_lips))
      if lip_form.is_valid():
        lip = lip_form.save()
        lip.root = l.get_root(lip.pattern, lip.inflection_characteristic)
        if lip.root is None:
          raise AjaxError(u'Niepasujące zakończenie formy podstawowej.')
        for qualifier in lip_form.fields['qualifiers'].queryset:
          qualifier.set_for(
            lip, qualifier in lip_form.cleaned_data['qualifiers'])
        lip.save()
      else:
        raise AjaxError(error_messages(lip_form))
    if name.startswith('cr_add') and prefix not in submitted_crs:
      submitted_crs.append(prefix)
      cr_form = CrossReferenceForm(data=form_dict, prefix=prefix)
      if cr_form.is_valid():
        cr_form.save()
      else:
        raise AjaxError(error_messages(cr_form))
  l.refresh_forms()
  if created:
    sort_rules = json.loads(request.session['sort_rules'])
    filters = json.loads(request.session['filters'])
    key = cache_key(
      sort_rules, filters, visible_vocabs, mask, filtering_mode(request.user))
    pk_list = cache.get(key)
    if pk_list:
      pk_list = [l.pk] + pk_list
      cache.set(key, pk_list)
  return {}

def update_lexeme_qualifiers(lexeme, vocabularies, rw_vocabularies, form_dict,
                             form_data):
  qualifiers = Qualifier.objects.filter(vocabulary__in=vocabularies)
  for qualifier in qualifiers:
    qualifier.set_for(
      lexeme, unicode(qualifier.pk) in form_dict['qualifiers'])
  for vocab in rw_vocabularies:
    set_vocabulary(lexeme, vocab, vocab.pk in form_dict['vocabularies'])
  submitted_lips = []
  for pair in form_data:
    name = pair['name']
    value = pair['value']
    prefix = name.split('-')[0]
    if name.startswith('lip') and prefix not in submitted_lips:
      submitted_lips.append(prefix)
      pk = int(prefix[3:])
      lip = LexemeInflectionPattern.objects.get(pk=pk)
      lip_qualifiers = get_list(form_data, prefix + '-qualifiers')
      for qualifier in qualifiers:
        qualifier.set_for(lip, unicode(qualifier.pk) in lip_qualifiers)
  return {}

def get_list(form_data, name):
  return [pair['value'] for pair in form_data if pair['name'] == name]

@ajax(method='post')
def delete_lexeme(request, id):
  id = int(id)
  l = Lexeme.objects.get(pk=id)
  if not l.perm(request.user, 'change'):
    raise AjaxError('access denied')
  l.deleted = True
  l.save()
  key_list = cache.get('key_list', [])
  for key in key_list:
    pk_list = cache.get(key)
    if pk_list and id in pk_list:
      pk_list.remove(id)
      cache.set(key, pk_list)
  return {}

@ajax(method='get')
def check_pos(request, pos_id, ic_id):
  ic = InflectionCharacteristic.objects.get(pk=ic_id)
  if InflectionCharacteristic.objects.filter(
       entry=ic.entry, part_of_speech__pk=pos_id):
    return {'answer': 'yes'}
  else:
    return {'answer': 'no'}

@ajax(method='get')
def check_pattern(request, pattern_name, ic_id):
  lips = LexemeInflectionPattern.objects.filter(
    inflection_characteristic__pk=ic_id, pattern__name=pattern_name)
  lips = lips.exclude(lexeme__status='cand', lexeme__deleted=True)
  if lips.exists():
    return {'answer': 'yes'}
  else:
    return {'answer': 'no'}

@ajax(method='get')
def get_ics(request, pos_id):
  return {'ics': list(InflectionCharacteristic.objects.filter(
                        part_of_speech__pk=pos_id).values_list('pk', 'entry'))}

START_ID = 1000000

@ajax(method='post')
def create_lexeme(request, vocab_id):
  owner = Vocabulary.objects.get(pk=vocab_id)
  if request.user not in owner.editors.all():
    raise AjaxError('access denied')
  next_id = Lexeme.objects.filter(
    pk__gte=START_ID).aggregate(Max('id'))['id__max']
  next_id = next_id + 1 if next_id else START_ID
  l = Lexeme()
  l.id = next_id
  l.homonym_number = 1 # żal
  l.part_of_speech = PartOfSpeech.objects.get(symbol='subst')
  l.owner_vocabulary = owner
  l.status = 'cand'
  l.responsible = request.user
  l.deleted = True # proste i genialne!
  l.save()
  LexemeAssociation.objects.create(lexeme=l, vocabulary=owner)
  return {'id': l.id}

@ajax(method='get', login_required=True)
def homonym_count(request, entry, lexeme_id):
  vocabs = visible_vocabularies(request.user)
  lexemes = Lexeme.objects.filter(
    deleted=False, entry=entry).exclude(pk=lexeme_id)
  return {'count': lexemes.count()}

@ajax(method='post')
def save_columns(request, col_model, col_names, remap):
  request.session['colModel'] = col_model
  request.session['colNames'] = col_names
  request.session['remap'] = remap
  return {}


@ajax(method='post')
def add_vocabulary(request, name):
  if not request.user.has_perm('dictionary.manage_vocabulary'):
    raise AjaxError('access denied')
  vocab = Vocabulary.objects.create(id=name)
  vocab.managers.add(request.user) #add
  return {}

@ajax(method='get')
def vocabulary_privileges(request, name):
  if not request.user.has_perm('dictionary.manage_vocabulary'):
    raise AjaxError('access denied')
  vocab = Vocabulary.objects.get(id=name)
  to_return = {}
  to_return['managers'] = list(vocab.managers.values_list('pk', flat=True))
  to_return['viewers'] = list(vocab.all_viewers().values_list('pk', flat=True))
  to_return['editors'] = list(vocab.editors.values_list('pk', flat=True))
  return to_return

@ajax(method='post')
def set_vocabulary_privilege(request, name, id, type, set):
  if not request.user.has_perm('dictionary.manage_vocabulary'):
    raise AjaxError('access denied')
  vocab = Vocabulary.objects.get(id=name)
  user = User.objects.get(pk=id)
  if type == 'view':
    related_manager = vocab.viewers
  elif type == 'change':
    related_manager = vocab.editors
  elif type == 'manage':
    related_manager = vocab.managers
  if set:
    related_manager.add(user) #add
  else:
    related_manager.remove(user) #add
  return {}
